package java.util.concurrent.atomic;

import java.util.Objects;
import java.lang.reflect.Modifier;
import java.security.PrivilegedActionException;
import sun.reflect.misc.ReflectUtil;
import java.security.AccessController;
import java.security.PrivilegedExceptionAction;
import java.lang.reflect.Field;
import sun.misc.Unsafe;
import java.util.function.IntBinaryOperator;
import java.util.function.IntUnaryOperator;
import sun.reflect.CallerSensitive;
import sun.reflect.Reflection;

public abstract class AtomicIntegerFieldUpdater<T>
{
    @CallerSensitive
    public static <U> AtomicIntegerFieldUpdater<U> newUpdater(Class<U> clazz, String s) {
        return new AtomicIntegerFieldUpdater.AtomicIntegerFieldUpdaterImpl(clazz, s, Reflection.getCallerClass());
    }
    
    protected AtomicIntegerFieldUpdater() {
    }
    
    public abstract boolean compareAndSet(T p0, int p1, int p2);
    
    public abstract boolean weakCompareAndSet(T p0, int p1, int p2);
    
    public abstract void set(T p0, int p1);
    
    public abstract void lazySet(T p0, int p1);
    
    public abstract int get(T p0);
    
    public int getAndSet(T t, int n) {
        int value;
        do {
            value = this.get(t);
        } while (!this.compareAndSet(t, value, n));
        return value;
    }
    
    public int getAndIncrement(T t) {
        int value;
        do {
            value = this.get(t);
        } while (!this.compareAndSet(t, value, value + 1));
        return value;
    }
    
    public int getAndDecrement(T t) {
        int value;
        do {
            value = this.get(t);
        } while (!this.compareAndSet(t, value, value - 1));
        return value;
    }
    
    public int getAndAdd(T t, int n) {
        int value;
        do {
            value = this.get(t);
        } while (!this.compareAndSet(t, value, value + n));
        return value;
    }
    
    public int incrementAndGet(T t) {
        int value;
        int n;
        do {
            value = this.get(t);
            n = value + 1;
        } while (!this.compareAndSet(t, value, n));
        return n;
    }
    
    public int decrementAndGet(T t) {
        int value;
        int n;
        do {
            value = this.get(t);
            n = value - 1;
        } while (!this.compareAndSet(t, value, n));
        return n;
    }
    
    public int addAndGet(T t, int n) {
        int value;
        int n2;
        do {
            value = this.get(t);
            n2 = value + n;
        } while (!this.compareAndSet(t, value, n2));
        return n2;
    }
    
    public final int getAndUpdate(T t, IntUnaryOperator intUnaryOperator) {
        int value;
        do {
            value = this.get(t);
        } while (!this.compareAndSet(t, value, intUnaryOperator.applyAsInt(value)));
        return value;
    }
    
    public final int updateAndGet(T t, IntUnaryOperator intUnaryOperator) {
        int value;
        int applyAsInt;
        do {
            value = this.get(t);
            applyAsInt = intUnaryOperator.applyAsInt(value);
        } while (!this.compareAndSet(t, value, applyAsInt));
        return applyAsInt;
    }
    
    public final int getAndAccumulate(T t, int n, IntBinaryOperator intBinaryOperator) {
        int value;
        do {
            value = this.get(t);
        } while (!this.compareAndSet(t, value, intBinaryOperator.applyAsInt(value, n)));
        return value;
    }
    
    public final int accumulateAndGet(T t, int n, IntBinaryOperator intBinaryOperator) {
        int value;
        int applyAsInt;
        do {
            value = this.get(t);
            applyAsInt = intBinaryOperator.applyAsInt(value, n);
        } while (!this.compareAndSet(t, value, applyAsInt));
        return applyAsInt;
    }
    
    private static final class AtomicIntegerFieldUpdaterImpl<T> extends AtomicIntegerFieldUpdater<T>
    {
        private static final Unsafe U;
        private final long offset;
        private final Class<?> cclass;
        private final Class<T> tclass;
        
        AtomicIntegerFieldUpdaterImpl(Class<T> tclass, String s, Class<?> currentClass) {
            Field f;
            int modifiers;
            try {
                f = (Field)AccessController.doPrivileged(new PrivilegedExceptionAction<Field>() {
                    @Override
                    public Field run() throws NoSuchFieldException {
                        return tclass.getDeclaredField(s);
                    }
                });
                modifiers = f.getModifiers();
                ReflectUtil.ensureMemberAccess(currentClass, tclass, null, modifiers);
                ClassLoader classLoader = tclass.getClassLoader();
                ClassLoader classLoader2 = currentClass.getClassLoader();
                if (classLoader2 != null && classLoader2 != classLoader && (classLoader == null || !isAncestor(classLoader, classLoader2))) {
                    ReflectUtil.checkPackageAccess(tclass);
                }
            }
            catch (PrivilegedActionException ex) {
                throw new RuntimeException(ex.getException());
            }
            catch (Exception cause) {
                throw new RuntimeException(cause);
            }
            if (f.getType() != Integer.TYPE) {
                throw new IllegalArgumentException("Must be integer type");
            }
            if (!Modifier.isVolatile(modifiers)) {
                throw new IllegalArgumentException("Must be volatile type");
            }
            this.cclass = ((Modifier.isProtected(modifiers) && tclass.isAssignableFrom(currentClass) && !isSamePackage(tclass, currentClass)) ? currentClass : tclass);
            this.tclass = tclass;
            this.offset = AtomicIntegerFieldUpdater.AtomicIntegerFieldUpdaterImpl.U.objectFieldOffset(f);
            return;
        }
        
        private static boolean isAncestor(ClassLoader parent, ClassLoader classLoader) {
            do {
                parent = parent.getParent();
                if (classLoader == parent) {
                    return true;
                }
                continue;
            } while (parent != null);
            return false;
        }
        
        private static boolean isSamePackage(Class<?> clazz, Class<?> clazz2) {
            return clazz.getClassLoader() == clazz2.getClassLoader() && Objects.equals(getPackageName(clazz), getPackageName(clazz2));
        }
        
        private static String getPackageName(Class<?> clazz) {
            String name = clazz.getName();
            int lastIndex = name.lastIndexOf(46);
            return (lastIndex != -1) ? name.substring(0, lastIndex) : "";
        }
        
        private final void accessCheck(T t) {
            if (!this.cclass.isInstance(t)) {
                this.throwAccessCheckException(t);
            }
        }
        
        private final void throwAccessCheckException(T t) {
            if (this.cclass == this.tclass) {
                throw new ClassCastException();
            }
            throw new RuntimeException(new IllegalAccessException("Class " + this.cclass.getName() + " can not access a protected member of class " + this.tclass.getName() + " using an instance of " + t.getClass().getName()));
        }
        
        @Override
        public final boolean compareAndSet(T o, int expected, int x) {
            this.accessCheck(o);
            return AtomicIntegerFieldUpdater.AtomicIntegerFieldUpdaterImpl.U.compareAndSwapInt(o, this.offset, expected, x);
        }
        
        @Override
        public final boolean weakCompareAndSet(T o, int expected, int x) {
            this.accessCheck(o);
            return AtomicIntegerFieldUpdater.AtomicIntegerFieldUpdaterImpl.U.compareAndSwapInt(o, this.offset, expected, x);
        }
        
        @Override
        public final void set(T o, int x) {
            this.accessCheck(o);
            AtomicIntegerFieldUpdater.AtomicIntegerFieldUpdaterImpl.U.putIntVolatile(o, this.offset, x);
        }
        
        @Override
        public final void lazySet(T o, int x) {
            this.accessCheck(o);
            AtomicIntegerFieldUpdater.AtomicIntegerFieldUpdaterImpl.U.putOrderedInt(o, this.offset, x);
        }
        
        @Override
        public final int get(T o) {
            this.accessCheck(o);
            return AtomicIntegerFieldUpdater.AtomicIntegerFieldUpdaterImpl.U.getIntVolatile(o, this.offset);
        }
        
        @Override
        public final int getAndSet(T o, int newValue) {
            this.accessCheck(o);
            return AtomicIntegerFieldUpdater.AtomicIntegerFieldUpdaterImpl.U.getAndSetInt(o, this.offset, newValue);
        }
        
        @Override
        public final int getAndAdd(T o, int delta) {
            this.accessCheck(o);
            return AtomicIntegerFieldUpdater.AtomicIntegerFieldUpdaterImpl.U.getAndAddInt(o, this.offset, delta);
        }
        
        @Override
        public final int getAndIncrement(T t) {
            return this.getAndAdd(t, 1);
        }
        
        @Override
        public final int getAndDecrement(T t) {
            return this.getAndAdd(t, -1);
        }
        
        @Override
        public final int incrementAndGet(T t) {
            return this.getAndAdd(t, 1) + 1;
        }
        
        @Override
        public final int decrementAndGet(T t) {
            return this.getAndAdd(t, -1) - 1;
        }
        
        @Override
        public final int addAndGet(T t, int n) {
            return this.getAndAdd(t, n) + n;
        }
        
        static {
            U = Unsafe.getUnsafe();
        }
    }
}
